/*
  Name: File Security System
  Copyright: ____
  Author: Huang Guan
  Date: 30-12-07 08:59
  Description: NT file system drvier
*/

#include "driver.h"
#include "fast_io.h"
#include "operation.h"
#include "file_node.h"

//Declaration
PDRIVER_OBJECT fs_drv;      //Driver Object Pointer
PDEVICE_OBJECT fs_dev;      //Filter Control Device Object (CDO)
PAGED_LOOKASIDE_LIST name_list;
DYNAMIC_FUNCTION_POINTERS dyn_funcs;
PDEVICE_LIST dev_list;      //the device list records the devices that has been attached.
extern FAST_IO_DISPATCH fast_io;
UNICODE_STRING ins_code = CONSTANT_UNICODE_STRING(L"[-= Insufficient Resources =-]");

//static functions
static void load_funcs ();
static void DriverUnload(IN PDRIVER_OBJECT drv_obj);
static NTSTATUS fs_dispatcher (IN PDEVICE_OBJECT dev_obj, IN PIRP irp);
static void FsNotification ( IN PDEVICE_OBJECT dev_obj, IN BOOLEAN FsActive );
static NTSTATUS attach_fsdev ( IN PDEVICE_OBJECT dev_obj, IN PNAME_CONTROL namectl );
static NTSTATUS enum_disk_dev ( IN PDEVICE_OBJECT sys_dev );
static NTSTATUS AttachToMountedDevice (IN PDEVICE_OBJECT dev_obj, IN PDEVICE_OBJECT new_obj );
static NTSTATUS AddToList( PDEVICE_OBJECT dev_obj );

//
NTSTATUS DriverEntry(IN PDRIVER_OBJECT drv_obj, IN PUNICODE_STRING reg_path)
{
	UNICODE_STRING dev_name, nt_name; 
	NTSTATUS status;
	int i;
	
	//load dynamic dyn_funcs
	load_funcs();
	//allocate list space.
	dev_list = (PDEVICE_LIST) ExAllocatePool( NonPagedPool, sizeof(DEVICE_LIST)*MAX_DEVICES );
	if( dev_list == NULL )
	{
		dprint("Can't allocate myList.\n");
		return FALSE;
	}
	RtlZeroMemory( dev_list, sizeof(DEVICE_LIST)*MAX_DEVICES );
	//init protection
	file_node_init();
	
	//save the object pointer
	fs_drv = drv_obj;
	//reg path
	// \REGISTRY\MACHINE\SYSTEM\ControlSet001\Services\[DEVICE_NAME]	
	//dprint("RegistryPath: %wZ\n", reg_path);
	//set unload function
	fs_drv->DriverUnload = DriverUnload;
	
	//dprint("dev_obj: %x\n", myDriverObject->dev_obj );  0
	//dprint("DriverExtension: %x\n", myDriverObject->DriverExtension );  pointer*
	//dprint("DriverName: %wZ\n", &(myDriverObject->DriverName) );	\Driver\[DEVICE_NAME]
	
	//NameLookup will use it
	ExInitializePagedLookasideList( &name_list,
	                            NULL,
	                            NULL,
	                            0,
	                            LOOKASIDE_SIZE,
	                            FLT_POOL_TAG_NAME_BUFFER,
	                            0 );
	
	//make CDO
	RtlInitUnicodeString( &dev_name, L"\\Device\\fsec" );
	RtlInitUnicodeString( &nt_name, L"\\DosDevices\\fsec" );
	status = IoCreateDevice( fs_drv, 
	                         0,                      //豸չϢ
	                         &dev_name,
	                         FILE_DEVICE_FSEC,   //this number is my randomized.
	                         0,
	                         TRUE,
	                         &fs_dev );
	if (!NT_SUCCESS( status )) {
        dprint( "DriverEntry: Error creating control device object \"%wZ\", status=%x\n",
    	   &dev_name, status );
        return status;
	}

	status = IoCreateSymbolicLink( &nt_name, &dev_name );
	if (!NT_SUCCESS( status )) {
        DbgPrint( "DriverEntry: Error IoCreateSymbolicLink \"%wZ\", status=%x\n",
        	&nt_name, status );
        IoDeleteDevice( fs_dev );
        return status;
	}
  
	//if failed, remember to delete the device created before.
	
	//Set MJ Dispath functions
	for( i = 0; i <= IRP_MJ_MAXIMUM_FUNCTION; i++ ) {
		fs_drv->MajorFunction[i] = fs_dispatcher;
	}
	fs_drv->MajorFunction[IRP_MJ_CREATE] = nt_create;
	fs_drv->MajorFunction[IRP_MJ_CLEANUP] = nt_cleanup;
	fs_drv->MajorFunction[IRP_MJ_CLOSE] = nt_close;
	fs_drv->MajorFunction[IRP_MJ_READ] = nt_read;
	fs_drv->MajorFunction[IRP_MJ_WRITE] = nt_write;
	fs_drv->MajorFunction[IRP_MJ_SET_INFORMATION] = nt_setinfo;
	fs_drv->MajorFunction[IRP_MJ_DIRECTORY_CONTROL] = nt_dirctl;
	fs_drv->MajorFunction[IRP_MJ_DEVICE_CONTROL] = nt_devctl;
	
	//Fast IO access
	fs_drv->FastIoDispatch = &fast_io;
	
	status = IoRegisterFsRegistrationChange( fs_drv, FsNotification );
	if (!NT_SUCCESS( status )) {
		dprint("DriverEntry: Error registering FS change notification, status=%08x\n", status );
		fs_drv->FastIoDispatch = NULL;
		IoDeleteDevice( fs_dev );
		return status;
	}
	
	//now work
	ClearFlag( fs_drv->Flags, DO_DEVICE_INITIALIZING );
	//dprint("Driver load successfully.\n");
	return STATUS_SUCCESS;
}

//uninstall the driver
void DriverUnload(IN PDRIVER_OBJECT drv_obj)
{
	int i;
	PDEVICE_OBJECT p_dev;
	PDEVICE_EXTENSION dev_ext;
	LARGE_INTEGER interval;
	UNICODE_STRING nt_name;
	if(drv_obj != fs_drv)
		return;
	
	prot_status( PS_IGNORE );
	IoUnregisterFsRegistrationChange( drv_obj, FsNotification );
	ExDeletePagedLookasideList( &name_list );
	
	for( i=0; dev_list[i].DeviceObject; i++ )
	{
		p_dev = dev_list[i].DeviceObject;
    	dev_ext = p_dev->DeviceExtension;
    	if (NULL != dev_ext) {
	    	dprint("Detaching %wZ\n", &dev_ext->NLExtHeader.DosName );
			IoDetachDevice( dev_ext->NLExtHeader.AttachedToDeviceObject );
	    	//ObDereferenceObject( devExt->NLExtHeader.AttachedToDeviceObject );
    	}
	}
	//wait for detaching
	interval.QuadPart = (2 * DELAY_ONE_SECOND);      //delay 2 seconds
	KeDelayExecutionThread( KernelMode, FALSE, &interval );
	for( i=0; dev_list[i].DeviceObject; i++ )
	{
		p_dev = dev_list[i].DeviceObject;
	    dev_ext = p_dev->DeviceExtension;
	    if (NULL != dev_ext) {
	    	NLCleanupDeviceExtensionHeader(&dev_ext->NLExtHeader);
	    }
	    IoDeleteDevice( p_dev );
        //ObDereferenceObject( pDev );
	}
	IoDeleteDevice( fs_dev );
	ExFreePool( dev_list );
	RtlInitUnicodeString( &nt_name, L"\\DosDevices\\fsec" );
	IoDeleteSymbolicLink( &nt_name );
	fs_drv->FastIoDispatch = NULL;
	//exit protection
	file_node_cleanup();
}

//IRP Dispatcher
NTSTATUS fs_dispatcher (IN PDEVICE_OBJECT dev_obj, IN PIRP irp)
{
	PIO_STACK_LOCATION pio_stack;
	//dprint("DefaultDispatch.\n");
	
	static PUCHAR irp_str[] = {
	    "IRP_MJ_CREATE",
	    "IRP_MJ_CREATE_NAMED_PIPE",
	    "IRP_MJ_CLOSE",
	    "IRP_MJ_READ",
	    "IRP_MJ_WRITE",
	    "IRP_MJ_QUERY_INFORMATION",
	    "IRP_MJ_SET_INFORMATION",
	    "IRP_MJ_QUERY_EA",
	    "IRP_MJ_SET_EA",
	    "IRP_MJ_FLUSH_BUFFERS",
	    "IRP_MJ_QUERY_VOLUME_INFORMATION",
	    "IRP_MJ_SET_VOLUME_INFORMATION",
	    "IRP_MJ_DIRECTORY_CONTROL",
	    "IRP_MJ_FILE_SYSTEM_CONTROL",
	    "IRP_MJ_DEVICE_CONTROL",
	    "IRP_MJ_INTERNAL_DEVICE_CONTROL",
	    "IRP_MJ_SHUTDOWN",
	    "IRP_MJ_LOCK_CONTROL",
	    "IRP_MJ_CLEANUP",
	    "IRP_MJ_CREATE_MAILSLOT",
	    "IRP_MJ_QUERY_SECURITY",
	    "IRP_MJ_SET_SECURITY",
	    "IRP_MJ_POWER",
	    "IRP_MJ_SYSTEM_CONTROL",
	    "IRP_MJ_DEVICE_CHANGE",
	    "IRP_MJ_QUERY_QUOTA",
	    "IRP_MJ_SET_QUOTA",
	    "IRP_MJ_PNP"
	};
	
    pio_stack = IoGetCurrentIrpStackLocation( irp );
    ASSERT(pio_stack->MajorFunction != IRP_MJ_POWER);
  
    IoSkipCurrentIrpStackLocation( irp );
  
    return IoCallDriver( ((PDEVICE_EXTENSION) dev_obj->DeviceExtension)->NLExtHeader.AttachedToDeviceObject,
                       irp );
}

//fs notify
void FsNotification ( IN PDEVICE_OBJECT dev_obj, IN BOOLEAN FsActive )
{
    PNAME_CONTROL namectl;
    PAGED_CODE();
    //Display the names of all the file system we are notified of
    namectl = NLGetAndAllocateObjectName( dev_obj,
                                          &name_list );

    if (namectl == NULL) {

        dprint( "SfFsNotification: Not attaching to %p \"%wZ\", insufficient resources.\n",
                       dev_obj, &dev_obj->DriverObject->DriverName );
        return;
    }

    //dprint( "FsNotification: %s %p \"%wZ\"\n",
    //               (FsActive) ? "Activating file system" : "Deactivating file system",
    //               dev_obj,
    //               &namectl->Name );
                   
    //  Handle attaching/detaching from the given file system.
    if (FsActive) {
        attach_fsdev( dev_obj, namectl );
    } else {
        //DetachFromFileSystemDevice( dev_obj );
    }
    NLFreeNameControl( namectl, &name_list );
}


//attach my driver to the file system
NTSTATUS attach_fsdev ( IN PDEVICE_OBJECT dev_obj, IN PNAME_CONTROL namectl )
{
    PDEVICE_OBJECT new_dev;
    PDEVICE_EXTENSION dev_ext;
    UNICODE_STRING tmp_str;
    NTSTATUS status;
    PNAME_CONTROL tmp_name;
    PAGED_CODE();
    
    //See if this is a file system type we care about.  If not, return.
    if (dev_obj->DeviceType!=FILE_DEVICE_DISK_FILE_SYSTEM) {
    	//dprint("%wZ is not a disk file system.\n", &namectl->Name );
        return STATUS_SUCCESS;
    }
    
    RtlInitUnicodeString( &tmp_str, L"\\FileSystem\\Fs_Rec" );

    tmp_name = NLGetAndAllocateObjectName( dev_obj->DriverObject,
                                         &name_list );

    if (tmp_name == NULL) {
        dprint( "AttachToFileSystemDevice:   Error retrieving name, may attach to FS recognizer, status=%08x\n",
                       STATUS_INSUFFICIENT_RESOURCES );

    } else if (RtlCompareUnicodeString( &tmp_name->Name,
                                        &tmp_str, TRUE ) == 0) {
        //  If it is a recognizer, don't attach
        NLFreeNameControl( tmp_name, &name_list );
        //dprint("Oh! it's a recognizer. %wZ\n", &tmp_name->Name );
        return STATUS_SUCCESS;
    }
    NLFreeNameControl( tmp_name, &name_list );
    
    //  Enumerate all the mounted devices that currently
    //  exist for this file system and attach to them.
    status = enum_disk_dev( dev_obj );

    return STATUS_SUCCESS;
}

//load dynamic functions
void load_funcs ( )
{
    UNICODE_STRING functionName;
    RtlZeroMemory( &dyn_funcs, sizeof( dyn_funcs ) );
    //  For each routine that we would want to use, lookup its address in the
    //  kernel or HAL.  If it is not present, that field in our global
    //  SpyDynamicFunctions structure will be set to NULL.
    RtlInitUnicodeString( &functionName, L"FsRtlRegisterFileSystemFilterCallbacks" );
    dyn_funcs.RegisterFileSystemFilterCallbacks = MmGetSystemRoutineAddress( &functionName );
    RtlInitUnicodeString( &functionName, L"IoAttachDeviceToDeviceStackSafe" );
    dyn_funcs.AttachDeviceToDeviceStackSafe = MmGetSystemRoutineAddress( &functionName );
    RtlInitUnicodeString( &functionName, L"IoEnumerateDeviceObjectList" );
    dyn_funcs.EnumerateDeviceObjectList = MmGetSystemRoutineAddress( &functionName );
    RtlInitUnicodeString( &functionName, L"IoGetLowerDeviceObject" );
    dyn_funcs.GetLowerDeviceObject = MmGetSystemRoutineAddress( &functionName );
    RtlInitUnicodeString( &functionName, L"IoGetDeviceAttachmentBaseRef" );
    dyn_funcs.GetDeviceAttachmentBaseRef = MmGetSystemRoutineAddress( &functionName );
    RtlInitUnicodeString( &functionName, L"IoGetDiskDeviceObject" );
    dyn_funcs.GetDiskDeviceObject = MmGetSystemRoutineAddress( &functionName );
    RtlInitUnicodeString( &functionName, L"IoGetAttachedDeviceReference" );
    dyn_funcs.GetAttachedDeviceReference = MmGetSystemRoutineAddress( &functionName );
    RtlInitUnicodeString( &functionName, L"RtlGetVersion" );
    dyn_funcs.GetVersion = MmGetSystemRoutineAddress( &functionName );
}

//ȡ豸
NTSTATUS GetBaseDeviceObjectName ( IN PDEVICE_OBJECT dev_obj, OUT PNAME_CONTROL Name )
{
    NTSTATUS status;
    PAGED_CODE();
    //  Get the base file system device object
    if(dyn_funcs.GetDeviceAttachmentBaseRef==NULL)
    	return FALSE;
    dev_obj = (dyn_funcs.GetDeviceAttachmentBaseRef)( dev_obj );
    //  Get the name of that object
    status = NLGetObjectName( dev_obj, Name );
    //  Remove the reference added by IoGetDeviceAttachmentBaseRef
    ObDereferenceObject( dev_obj );
    return status;
}

//жǷǾ
NTSTATUS IsShadowCopyVolume (IN PDEVICE_OBJECT stack_dev, OUT PBOOLEAN IsShadowCopy )
{
  //PAGED_CODE();
  //Default to NOT a shadow copy volume
  UNICODE_STRING volSnapDriverName;
  WCHAR buffer[64];
  PUNICODE_STRING name;
  ULONG returnedLength;
  NTSTATUS status;
  
  //  In Windows XP, all ShadowCopy devices were of type FILE_DISK_DEVICE.
  //  If this does not have a device type of FILE_DISK_DEVICE, then
  //  it is not a ShadowCopy volume.  Return now.
  if (FILE_DEVICE_DISK != stack_dev->DeviceType) {
      return STATUS_SUCCESS;
  }
  //  Unfortunately, looking for the FILE_DEVICE_DISK isn't enough.  We
  //  need to find out if the name of this driver is \Driver\VolSnap as
  //  well.
  name = (PUNICODE_STRING) buffer;
  RtlInitEmptyUnicodeString( name,
                         		Add2Ptr( name, sizeof( UNICODE_STRING ) ),
                         		sizeof( buffer ) - sizeof( UNICODE_STRING ) );
  status = ObQueryNameString( stack_dev,
                              (POBJECT_NAME_INFORMATION)name,
                              name->MaximumLength,
                              &returnedLength );
  if (!NT_SUCCESS( status )) {
    return status;
  }
  RtlInitUnicodeString( &volSnapDriverName, L"\\Driver\\VolSnap" );
  if (RtlEqualUnicodeString( name,
                             &volSnapDriverName, TRUE )) {
    *IsShadowCopy = TRUE;
  } else {
    NOTHING;
  }
  return STATUS_SUCCESS;
}



//enumerate storage file systems and add to list
NTSTATUS enum_disk_dev ( IN PDEVICE_OBJECT sys_dev )
{
    PDEVICE_OBJECT new_dev;
    PDEVICE_EXTENSION dev_ext;
    PDEVICE_OBJECT *devs;
    PDEVICE_OBJECT stack_dev;
    PNAME_CONTROL name;
    NTSTATUS status;
    ULONG nr_devs;
    ULONG i;
    BOOLEAN isShadowCopyVolume;
    BOOLEAN hasLock = FALSE;

    PAGED_CODE();
    //  Find out how big of an array we need to allocate for the
    //  mounted device list.
    status = (dyn_funcs.EnumerateDeviceObjectList)(
                    sys_dev->DriverObject,
                    NULL,
                    0,
                    &nr_devs);
    //  We only need to get this list of there are devices.  If we
    //  don't get an error there are no devices so go on.
    if (STATUS_BUFFER_TOO_SMALL != status) {
        return status;
    }

    //  Allocate memory for the list of known devices
    nr_devs += 8;        //grab a few extra slots

    devs = ExAllocatePoolWithTag( NonPagedPool,
                                     (nr_devs * sizeof(PDEVICE_OBJECT)),
                                     FLT_POOL_TAG_ENUMFSVOL );
    if (NULL == devs) {
        return STATUS_INSUFFICIENT_RESOURCES;
    }
	//dprint("Getting Device Object List ...\n");
    //  Now get the list of devices.  If we get an error again
    //  something is wrong, so just fail.
    status = (dyn_funcs.EnumerateDeviceObjectList)(
                    sys_dev->DriverObject,
                    devs,
                    (nr_devs * sizeof(PDEVICE_OBJECT)),
                    &nr_devs);

    if (!NT_SUCCESS( status ))  {
        ExFreePoolWithTag( devs, FLT_POOL_TAG_ENUMFSVOL );
        return status;
    }
    //  Allocate the name control structure.  We'll use this same name
    //  buffer each time through the for loop.
    status = NLAllocateNameControl( &name, &name_list );

    if (!NT_SUCCESS( status )) {
        //  If we couldn't get it then we can not process any of the
        //  entries, release all of the device objects and return.
        for (i=0; i<nr_devs; i++) {
            ObDereferenceObject( devs[i] );
        }
        ExFreePoolWithTag( devs, FLT_POOL_TAG_ENUMFSVOL );
        goto SkipAttach;
    }

    //  Walk the given list of devices and attach to them if we should.
    for (i=0; i < nr_devs; i++) {
        //  Initialize state so it will look like a clean name each time
        name->Name.Length = 0;
        stack_dev = NULL;
        new_dev = NULL;
        isShadowCopyVolume = FALSE;
        try {
            //  Do not attach if:
            //      - This is the control device object (the one passed in)
            //      - The device type does not match
            //      - We are already attached to it.
            if ((devs[i] == sys_dev) ||
                (devs[i]->DeviceType != sys_dev->DeviceType) 
                // check if we've already attached  :: not implemented.
                ) {
                leave;
            }
            //  See if this device has a name.  If so, then it must
            //  be a control device so don't attach to it.  This handles
            //  drivers with more then one control device (like FastFat).
            //  We also don't want to attach if we could not get a name.
            status = GetBaseDeviceObjectName( devs[i], name );

            if (!NT_SUCCESS(status) || (name->Name.Length > 0)) {
                dprint("this device has name: %wZ\n", &name->Name);
                leave;
            }
            //  Get the real (disk/storage stack) device object associated
            //  with this file system device object.  Only try to attach
            //  if we have a disk/storage stack device object.
            status = (dyn_funcs.GetDiskDeviceObject)( devs[i],
                         &stack_dev );

            if (!NT_SUCCESS( status )) {
            	dprint("Failed......\n");
                stack_dev = NULL;
                leave;
            }
            //  Determine if this is a shadow copy volume.  If so don't
            //  attach to it.
            status = IsShadowCopyVolume ( stack_dev,
                                            &isShadowCopyVolume );

            if (NT_SUCCESS(status) &&
                isShadowCopyVolume ) {
                PNAME_CONTROL shadowDeviceName;
                //  Get the name for the debug display
                shadowDeviceName = NLGetAndAllocateObjectName( stack_dev,
                                                               &name_list );

                dprint( "EnumerateFileSystemVolumes Not attaching to Volume  %p \"%wZ\", shadow copy volume\n",
                               stack_dev,
                               shadowDeviceName ? &shadowDeviceName->Name :
                                                  &ins_code );

                if (shadowDeviceName != NULL) {
                    NLFreeNameControl( shadowDeviceName, &name_list );
                }
                leave;
            }
            //  Allocate a new device object to attach with
            status = IoCreateDevice( fs_drv,
                                     sizeof( DEVICE_EXTENSION ),
                                     NULL,
                                     devs[i]->DeviceType,
                                     0,
                                     FALSE,
                                     &new_dev );
            if (!NT_SUCCESS( status )) {
                new_dev = NULL;
                dprint("create device failed.\n");
                leave;
            }
            //  Initialize the device extension.
            dev_ext = new_dev->DeviceExtension;
            dev_ext->Flags = 0;
            NLInitDeviceExtensionHeader( &dev_ext->NLExtHeader,
                                         new_dev,
                                         stack_dev );
            //  Set storage stack device name.  Just reuse the name
            //  NAME_CONTROL.
            name->Name.Length = 0;
            status = NLGetObjectName( stack_dev, name );
            if (!NT_SUCCESS( status)) {
            	dprint("GetOName failed.\n");
                leave;
            }
            
            if( stack_dev==devs[i])
            	dprint("!!!! Oh~~~They are equal ==== \n");
            //  Copy it to the device extension.  We allocate from non-paged
            //  pool so this memory is always available for debugging (never
            //  gets paged out).
            status = NLAllocateAndCopyUnicodeString( &dev_ext->NLExtHeader.DeviceName,
                                                     &name->Name,
                                                     FLT_POOL_TAG_DEVNAME );
            if (!NT_SUCCESS(status)) {
            	dprint("Copy failed.\n");
                leave;
            }
            status = AttachToMountedDevice( devs[i], new_dev );
            if (!NT_SUCCESS( status )) {
            	dprint("AttachToMountedDevice failed.\n");
            	leave;
            }
            status = RtlVolumeDeviceToDosName( stack_dev,
                &dev_ext->NLExtHeader.DosName);
    		if (!NT_SUCCESS( status )) {
        		dev_ext->NLExtHeader.DosName.Length = 0;
			}
            if(NT_SUCCESS( status ))
				dprint("Device: %wZ  Dos Name: %wZ\n", &name->Name, &dev_ext->NLExtHeader.DosName );		
            AddToList( new_dev );
            //  ObDereferenceObject( storageStackDeviceObject );
            if (stack_dev!=devs[i]) {
              ObDereferenceObject( stack_dev );
              //dprint("ObDereferenceObject storageStackDeviceObject\n");
            }
            //  Mark not to free the device object
            new_dev = NULL;
        } finally {
        	/*
            if (stack_dev != NULL && stack_dev!=devs[i]) {
              ObDereferenceObject( stack_dev );
            }*/
            //  Cleanup the device object if still defined
            if (new_dev != NULL) {
          		NLCleanupDeviceExtensionHeader(&dev_ext->NLExtHeader);
                IoDeleteDevice( new_dev );
            }
            //EnumerateDeviceObjectList increase the reference of devs[i]
            ObDereferenceObject( devs[i] );
        }
    }
    NLFreeNameControl( name, &name_list );
SkipAttach:
    status = STATUS_SUCCESS;
    ExFreePoolWithTag( devs, FLT_POOL_TAG_ENUMFSVOL );
    return status;
}

//attach to a mounted device: detailed thing will do
NTSTATUS AttachToMountedDevice (IN PDEVICE_OBJECT dev_obj, IN PDEVICE_OBJECT new_dev )
{
	PDEVICE_EXTENSION dev_ext = new_dev->DeviceExtension;
	NTSTATUS status;
	ULONG i;
	PAGED_CODE();
  
	if (FlagOn( dev_obj->Flags, DO_BUFFERED_IO )) {
	   SetFlag( new_dev->Flags, DO_BUFFERED_IO );
	}
	if (FlagOn( dev_obj->Flags, DO_DIRECT_IO )) {
		SetFlag( new_dev->Flags, DO_DIRECT_IO );
  	}
	for (i=0; i < 8; i++) {
	    LARGE_INTEGER interval;
	    /*
	    status = AttachDeviceToDeviceStack( new_obj,
	                                 dev_obj,
	                                 &dev_ext->NLExtHeader.AttachedToDeviceObject );*/
	    status = (dyn_funcs.AttachDeviceToDeviceStackSafe)( new_dev,
		            dev_obj,
		            &dev_ext->NLExtHeader.AttachedToDeviceObject );
	    if (NT_SUCCESS(status)) {
	        ClearFlag( new_dev->Flags, DO_DEVICE_INITIALIZING );
	        //DbgPrint( "AttachToMountedDevice: Attaching %p to volume %p \"%wZ\"\n",
	        //               new_dev,
	        //               dev_ext->NLExtHeader.AttachedToDeviceObject,
	        //               &dev_ext->NLExtHeader.DeviceName );
	
	        return STATUS_SUCCESS;
	    }
	    //  Delay, giving the device object a chance to finish its
	    //  initialization so we can try again
	    interval.QuadPart = (500 * DELAY_ONE_MILLISECOND);
		KeDelayExecutionThread( KernelMode, FALSE, &interval );
	}
	return status;
}

//add the attached device to list
NTSTATUS AddToList( PDEVICE_OBJECT dev_obj )
{
	int i;
	for( i=0; i<MAX_DEVICES; i++ )
	{
		if( dev_list[i].DeviceObject==NULL ){
			dev_list[i].DeviceObject = dev_obj;
			return STATUS_SUCCESS;
		}
	}
	dprint("ERROR; myList is full.\n");
	return FALSE;
}

