#ifndef _DRIVER_H_
#define _DRIVER_H_

////////////////////////////////////////////////////////////////
//FS_VERSION
#define FS_VERSION "File Security System [2008-1-1](0.9)"

/// NT Driver  ///
#include <ntifs.h>
#include <namelookup.h>
#include "ioctl.h"

#define dprint DbgPrint
#define ENTER_FUNCTION() 

#define MAX_DEVICES 32

#ifndef SetFlag
#define SetFlag(_F,_SF)       ((_F) |= (_SF))
#endif

#ifndef ClearFlag
#define ClearFlag(_F,_SF)     ((_F) &= ~(_SF))
#endif

#ifndef Add2Ptr
#define Add2Ptr(P,I) ((PVOID)((PUCHAR)(P) + (I)))
#endif

#define IS_MY_DEVICE_OBJECT(_devObj) \
    (((_devObj) != NULL) && \
     ((_devObj)->DriverObject == fs_drv) && \
      ((_devObj)->DeviceExtension != NULL))
#define IS_MY_CONTROL_DEVICE_OBJECT(_devObj) \
    (((_devObj) == fs_dev) ? \
            (ASSERT(((_devObj)->DriverObject == fs_drv) && \
                    ((_devObj)->DeviceExtension == NULL)), TRUE) : \
            FALSE)

#define LOOKASIDE_SIZE  sizeof( NAME_CONTROL )
#define FLT_POOL_TAG_NAME_BUFFER   'bnYM'
#define FLT_POOL_TAG_DEVNAME       'nvYM'
#define FLT_POOL_TAG_ENUMFSVOL     'neYM'

#define DELAY_ONE_MICROSECOND   (-10)
#define DELAY_ONE_MILLISECOND   (DELAY_ONE_MICROSECOND*1000)
#define DELAY_ONE_SECOND        (DELAY_ONE_MILLISECOND*1000)

#define CONSTANT_UNICODE_STRING(s)   { sizeof( s ) - sizeof( WCHAR ), sizeof(s), s }

typedef struct _DEVICE_EXTENSION {
	//For looking up name
  NL_DEVICE_EXTENSION_HEADER NLExtHeader;
  //Local flags for this device
  ULONG Flags;
} DEVICE_EXTENSION, *PDEVICE_EXTENSION;
typedef struct _DEVICE_LIST{
	PDEVICE_OBJECT DeviceObject;	//AttachedDO, չϢ
}DEVICE_LIST, *PDEVICE_LIST;

typedef
NTSTATUS
(*P_REGISTER_FILE_SYSTEM_FILTER_CALLBACKS) (
    IN PDRIVER_OBJECT DriverObject,
    IN PFS_FILTER_CALLBACKS Callbacks
    );
typedef
NTSTATUS
(*P_ENUMERATE_DEVICE_OBJECT_LIST) (
    IN  PDRIVER_OBJECT DriverObject,
    IN  PDEVICE_OBJECT *DeviceObjectList,
    IN  ULONG DeviceObjectListSize,
    OUT PULONG ActualNumberDeviceObjects
    );
typedef
NTSTATUS
(*P_ATTACH_DEVICE_TO_DEVICE_STACK_SAFE) (
    IN PDEVICE_OBJECT SourceDevice,
    IN PDEVICE_OBJECT TargetDevice,
    OUT PDEVICE_OBJECT *AttachedToDeviceObject
    );
typedef
PDEVICE_OBJECT
(*P_GET_LOWER_DEVICE_OBJECT) (
    IN  PDEVICE_OBJECT  DeviceObject
    );
typedef
PDEVICE_OBJECT
(*P_GET_DEVICE_ATTACHMENT_BASE_REF) (
    IN PDEVICE_OBJECT DeviceObject
    );
typedef
NTSTATUS
(*P_GET_DISK_DEVICE_OBJECT) (
    IN  PDEVICE_OBJECT  FileSystemDeviceObject,
    OUT PDEVICE_OBJECT  *DiskDeviceObject
    );
typedef
PDEVICE_OBJECT
(*P_GET_ATTACHED_DEVICE_REFERENCE) (
    IN PDEVICE_OBJECT DeviceObject
    );
typedef
NTSTATUS
(*P_GET_VERSION) (
    IN OUT PRTL_OSVERSIONINFOW VersionInformation
    );
typedef struct _DYNAMIC_FUNCTION_POINTERS {
    //  The following routines should all be available on Windows XP (5.1) and
    //  later.
    P_REGISTER_FILE_SYSTEM_FILTER_CALLBACKS RegisterFileSystemFilterCallbacks;
    P_ATTACH_DEVICE_TO_DEVICE_STACK_SAFE AttachDeviceToDeviceStackSafe;
    P_ENUMERATE_DEVICE_OBJECT_LIST EnumerateDeviceObjectList;
    P_GET_LOWER_DEVICE_OBJECT GetLowerDeviceObject;
    P_GET_DEVICE_ATTACHMENT_BASE_REF GetDeviceAttachmentBaseRef;
    P_GET_DISK_DEVICE_OBJECT GetDiskDeviceObject;
    P_GET_ATTACHED_DEVICE_REFERENCE GetAttachedDeviceReference;
    P_GET_VERSION GetVersion;
} DYNAMIC_FUNCTION_POINTERS, *PDYNAMIC_FUNCTION_POINTERS;


/////////////////////////////NT DRIVER///////////////////////////////


extern PDRIVER_OBJECT fs_drv;      //Driver Object Pointer
extern PDEVICE_OBJECT fs_dev;      //Filter Control Device Object (CDO)
extern PAGED_LOOKASIDE_LIST name_list;;
#endif //_DRIVER_H_
