/*
  Name: File Security System
  Copyright: ____
  Author: Huang Guan
  Date: 30-12-07 08:59
  Description: NT Process operations
*/

#include "driver.h"
#include "process.h"
#include "util.h"
#include "rlist.h"

#define MAX_PATH_LEN 1024
#define MAX_SYS_PATH 320

struct sysproc_list{
	int count;
	wchar_t list[MAX_SYS_PATH][MAX_PATH_LEN];
};

static struct sysproc_list * sp_list = NULL;

typedef struct su_entry_s{
	wchar_t name[MAX_PATH_LEN];
}su_entry_t;
rlist_t	su_list;
int mut_su_list, mut_sp_list;

//ȡļ· 
wchar_t* process_file_path()
{
	unsigned int dwAddress = (unsigned int)PsGetCurrentProcess();
	wchar_t* name;
	if(dwAddress == 0 || dwAddress == 0xFFFFFFFF)
		return NULL;
	dwAddress += 0x1B0;
	if((dwAddress = *(unsigned int*)dwAddress) == 0) return 0;
	dwAddress += 0x10;
	if((dwAddress = *(unsigned int*)dwAddress) == 0) return 0;
	dwAddress += 0x3C;
	if((dwAddress = *(unsigned int*)dwAddress) == 0) return 0;
	name = (wchar_t*)dwAddress;
	if( _wcsnicmp( name, L"\\??\\", 4 ) == 0 )
		name += 4;
	return name;
} 

void su_remover( int v )
{
	su_entry_t *t = (su_entry_t*) v;
	kmem_free( t );
}

int su_comparer( int v1, int v2 )
{
	wchar_t *input = (wchar_t*) v1;
	su_entry_t *t = (su_entry_t*) v2;
	return _wcsnicmp( input, t->name, MAX_PATH_LEN );
}

int process_init()
{
	sp_list = (struct sysproc_list*)kmem_get( sizeof(struct sysproc_list));
	if( sp_list == NULL )
	{
		dprint("Allocate memory failed.");
		return -1;
	}
	sp_list->count = 0;
	//super process
	rlist_init( &su_list, 32, su_remover, su_comparer );
	mut_init( &mut_su_list );
	mut_init( &mut_sp_list );
	return 0;
}

void process_cleanup()
{
	mut_down( &mut_sp_list );
	kmem_free( sp_list );
	sp_list = NULL;
	mut_up( &mut_sp_list );
	//
	mut_down( &mut_su_list );
	rlist_cleanup( &su_list );
	mut_up( &mut_su_list );
}

int process_add_sys( wchar_t* p )
{
	//Warning: Avoid race condition.
	//Only one thread can run this code at the moment.
	int i;
	mut_down( &mut_sp_list );
	if( !sp_list || !p ) {
		return -1;
		mut_up( &mut_sp_list );
	}
	i = sp_list->count;
	if( i>= MAX_SYS_PATH ) {
		mut_up( &mut_sp_list );
		return -2;
	}
	wcsncpy( sp_list->list[i], p, MAX_PATH_LEN );
	sp_list->count ++;
	mut_up( &mut_sp_list );
	return 0;
}

int process_set_sys( char* buf, int size )
{
	int i;
	if( size > sizeof(struct sysproc_list) )
		return -1;
	mut_down( &mut_sp_list );
	if( !sp_list || !buf ) {
		return -1;
		mut_up( &mut_sp_list );
	}
	RtlCopyBytes( sp_list, buf, size );
	for( i=0; i<sp_list->count; i++ )
		dprint("ProcList[%d]: %ws", i, sp_list->list[i] );
	mut_up( &mut_sp_list );
	dprint("##Update Process List.");
	return 0;
}

int process_add_super( wchar_t* p )
{
	su_entry_t* t;
	if( p == NULL ) return -1;
	t = (su_entry_t*)kmem_get( sizeof( su_entry_t ) );
	if( t == NULL ) return -1;
	wcsncpy( t->name, p, MAX_PATH_LEN );
	mut_down( &mut_su_list );
	if( su_list.list )
		rlist_append( &su_list, (int)t );
	else
		kmem_free( t );
	mut_up( &mut_su_list );
	return 0; 
}

int process_is_super( )
{
	wchar_t* p = process_file_path();
	if( p == NULL ) return 1; //if p==NULL, maybe a special process
	mut_down( &mut_su_list );
	if( rlist_find( &su_list, (int)p ) )
	{
		dprint("Aha! Super Process!");
		mut_up( &mut_su_list );
		return 1;
	}
	mut_up( &mut_su_list );
	return 0;
}

static int process_match_sys( wchar_t* p )
{
	int i;
	mut_down( &mut_sp_list );
	if( !sp_list ) return -1;
	for( i=0; i<sp_list->count; i++ )
	{
		int len, j;
		len = wcslen( sp_list->list[i] );
		if( _wcsnicmp( sp_list->list[i], p, len ) == 0 )
		{
			mut_up( &mut_sp_list );
			return 1; //matched.
		}
	}
	mut_up( &mut_sp_list );
	return 0;
}

int process_is_sys()
{
	wchar_t* proc_file = process_file_path();
	if( proc_file == NULL )
		return 1;	//I don't know, return 1.
	//proc_file: C:\WINDOWS\system32\cmd.exe
	return process_match_sys( proc_file );
}
